(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright (c) 2013-2014 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: display_quantity_components.ml,v 1.18 2014/05/06 12:56:21 fclement Exp $ *)

module Nat = Unatural;;
module Prational = Upositive_rational;;

module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _2_ = Nat._2_
and _3_ = Nat._3_
;;

let natural_axpby a x b y = Nat.add (Nat.mul a x) (Nat.mul b y);;

(* Grimage shortcuts. *)

let blank = 0;; (* = Graphics.black *)

let pixmap_make (w, h) c = Umatrix.to_array_array (Umatrix.make h w c);;

let pixmap_draw pxm (x, y) =
  Pixmap.draw_pixmap pxm (Nat.to_int x) (Nat.to_int y)
;;

open Refinement_types;;
open Display_types;;
open Display_types_print;;
open Scalar_spec;;
open Components_spec;;
open Display_window;;

(* Lower-left corner basics. *)

let lower_left_corner_make (nx, ny) m (ax, bx, ay, by) =
  let two_ms = Nat.mul _2_ m in
  let sub_hor = Nat.add nx two_ms
  and sub_ver = Nat.add ny two_ms in
  let ll_x = natural_axpby ax sub_hor bx m
  and ll_y = natural_axpby ay sub_ver by m in
  (ll_x, ll_y)
;;

let abxy ax ay = (ax, _1_, ay, _1_);;

let get_abxy_p compl = function
  | (No_graphics, _) -> assert false
  | (Parameter_graphics, Horizontal_comps_dir) -> Some (fun i -> abxy i _0_)
  | (Full_graphics, Horizontal_comps_dir) -> Some (fun i -> abxy i _1_)
  | (_, Vertical_comps_dir) -> Some (fun i -> abxy _0_ (compl i))
;;

let get_abxy_g compl = function
  | (No_graphics, _) -> assert false
  | (Parameter_graphics, _) -> None
  | (Full_graphics, Horizontal_comps_dir) -> Some (fun i -> abxy i _0_)
  | (Full_graphics, Vertical_comps_dir) -> Some (fun i -> abxy _1_ (compl i))
;;

let get_abxy = function
  | Param -> get_abxy_p
  | Grad -> get_abxy_g
;;

(* Print pixmap basics. *)

let component_name i = Printf.sprintf "comp #%d" (Nat.to_int i);;

(* Alter pixmap basics. *)

let set_pxm_pxl pxm c i j = pxm.(i).(j) <- c;;

(* Function scale is the multiplication by a rational number scaling factor.
 It transforms coordinates in the computed image into coordinates in the
 displayed image. Therefore, the rational coefficient is greater than 1 when
 the displayed image is larger than the computed image, and of course it is
 lesser than 1 otherwise. *)

let set_image_pixel set_pixel =
  let { scaling_factor = factor; _; } = get_graphics_elements () in
  let scale = Prational.scale factor in
  (fun im_i im_j ->
     let imin = Nat.to_int (scale im_i)
     and imax = pred (Nat.to_int (scale (Nat.succ im_i))) in
     let jmin = Nat.to_int (scale im_j)
     and jmax = pred (Nat.to_int (scale (Nat.succ im_j))) in
     for i = imin to imax do
       for j = jmin to jmax do
         set_pixel i j;
       done;
     done;)
;;

let set_zone_pixel set_pixel stc imgs dom idx _uidx =
  let x =
    match imgs with
    | Mapping.Images ims -> ims.(Uindex.to_int idx)
    | Mapping.Constant x -> x in
  let c = stc x in
  let set_pixel = set_pixel c in
  let (im_i, im_j) = Set.element dom idx in
  set_image_pixel set_pixel im_i im_j;
;;

let set_zone pxm stc _zn { Mapping.domain = dom; images = imgs; } =
  let set_pixel = set_pxm_pxl pxm in
  let set_zone_pixel = set_zone_pixel set_pixel stc imgs dom in
  Set.iter set_zone_pixel dom;
;;

module Make =
  functor (Scal : SCALAR) ->
  functor (Comps : COMPONENTS) ->
struct
  module H = History.Make (Scal) (Comps);;
  module Bb = H.Bb;;
  module St = H.St;;
  
  module Dqt = Display_quantity_types.Make (Scal) (Comps);;
  
  let compl =
    let n = Comps.get_dim () in
    (fun i -> Nat.sub n i)
  ;;
  
  let blank = blank;;
  
  let fk_comps = Comps.make;;
  
  let assert_false = (fun _ -> assert false);;
  let assert_false_comps = fk_comps assert_false;;
  
  let mk_pq _fake comps _gm = comps;;
  
  let mk_gq fake comps = function
    | No_graphics -> assert false
    | Parameter_graphics -> fake
    | Full_graphics -> comps
  ;;
  
  let mk_xq = function
    | Param -> mk_pq
    | Grad -> mk_gq
  ;;
  
  let mk_comps q fk = (mk_xq q) (fk_comps fk);;
  
  let get_comps comps = Comps.map_rails comps (Comps.make ());;
  
  (* For late enough evaluation of its components, it is easier to only consider
   functional comps. If necessary, thunks are added, i.e. for pixmaps, and
   lower_left_corners.  *)
   
  (* Pixmaps. *)
  
  let fk_get_pxm = (fun () -> [||]);;
  
  let mk_get_pxm sz = (fun () -> pixmap_make sz blank);;
  let mk_get_pxm_comps sz = Comps.make (mk_get_pxm sz);;
  
  let mk_get_pxms q gm sz =
    let fk = fk_get_pxm
    and comps = mk_get_pxm_comps sz in
    mk_comps q fk comps gm
  ;;
  
  let make_pixmaps q gm sz = get_comps (mk_get_pxms q gm sz);;
  
  (* Lower_left_corners. *)
  
  let fk_get_ll = (fun () -> (_0_, _0_));;
  
  let mk_get_ll sz m abxy = (fun () -> lower_left_corner_make sz m abxy);;
  let mk_get_ll_comps q gm comps_dir sz m =
    match get_abxy q compl (gm, comps_dir) with
    | None -> assert_false_comps
    | Some get_abxy -> Comps.init (fun i -> mk_get_ll sz m (get_abxy i))
  ;;
  
  let mk_get_lls q gm comps_dir sz m =
    let fk = fk_get_ll
    and comps = mk_get_ll_comps q gm comps_dir sz m in
    mk_comps q fk comps gm
  ;;
  
  let make_lower_left_corners q gm comps_dir sz m =
    get_comps (mk_get_lls q gm comps_dir sz m)
  ;;
  
  (* Print_pixmaps. *)
  
  let fk_prt = (fun _ppf -> ());;
  
  let fake_print_pixmaps = fk_comps fk_prt;;
  
  let mk_prt q s pxm ppf =
    Format.fprintf ppf "@[<hv 2>%a %s pxm =@ %a@]"
      print_quantity q s print_pixmap pxm
  ;;
  let mk_prt_comps q = Comps.map2 (mk_prt q) (Comps.init component_name);;
  
  let make_print_pixmaps q pxms gm =
    let fk = fk_prt
    and comps = mk_prt_comps q pxms in
    mk_comps q fk comps gm
  ;;
  
  (* Alter_pixmaps. *)
  
  let fk_alt = (fun _iter _stc _pp -> ());;
  
  let fake_alter_pixmaps = fk_comps fk_alt;;
  
  let mk_alt pxm iter stc pp = iter (set_zone pxm stc) pp;;
  let mk_alt_comps pxms = Comps.map mk_alt pxms;;
  
  let make_alter_pixmaps q pxms gm =
    let fk = fk_alt
    and comps = mk_alt_comps pxms in
    mk_comps q fk comps gm
  ;;
  
  (* Draw_pixmaps. *)
  
  let fk_drw = (fun () -> ());;
  
  let fake_draw_pixmaps = fk_comps fk_drw;;
  
  let mk_drw pxm ll = (fun () -> pixmap_draw pxm ll);;
  let mk_drw_comps = Comps.map2 mk_drw;;
  
  let make_draw_pixmaps q pxms lls gm =
    let fk = fk_drw
    and comps = mk_drw_comps pxms lls in
    mk_comps q fk comps gm
  ;;
  
  (* Application. *)
  
  let apply_comps apply rc = Comps.iter apply rc;;
  
  let print_apply print ppf = apply_comps (fun f -> print f ppf)
  and draw_apply = apply_comps (fun f -> f ())
  ;;
  
  let alter_apply iter stcs pps comps =
    let comps = Comps.map (fun f -> f iter) comps in
    Comps.iter_rails2 comps stcs pps;
  ;;
end
;;
