(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright (c) 2013-2014 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: wrapper.ml,v 1.14 2014/06/25 14:55:52 fclement Exp $ *)

module Nat = Unatural;;

let format_sprint print ident x =
  let ppf = Format.str_formatter in
  Format.fprintf ppf "@[<hv 4>@ %s =@ %a@]" ident print x;
  Format.flush_str_formatter ()
;;

open Wire_types_print;;

let w_qual = Upervasives.qualify "Wrapper";;

let w_failwith = Upervasives.failwith_qual w_qual;;

(* Init. *)

let mk_init_inj { Udebug.spf_report = report; spf_debug = debug; _; } dim =
  report "sending one value";
  report "sending dimension";
  debug (format_sprint print_dimension "dim" dim);
  [| Typed_value.of_int (Nat.to_int dim); |]
;;

let mk_init_proj { Udebug.spf_report = report; spf_debug = debug; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._5_ then
    w_failwith "init_proj" "initialization should return five values" else
  report "five values received";
  let (ni, nj) = Typed_value.to_int_couple vals.(0) in
  let sz = (Nat.of_int ni, Nat.of_int nj) in
  report "domain size extracted";
  debug (format_sprint print_size "sz" sz);
  let ppm = Typed_value.to_c_int_matrix vals.(1) in
  let ppm = Umatrix.map Nat.of_int ppm in
  report "initial partition extracted";
  debug (format_sprint print_partition "ppm" ppm);
  let lens = Typed_value.to_int_vector vals.(2) in
  let lens = Array.map Nat.of_int lens in
  report "initial coarse parameter lengths extracted";
  let a = Typed_value.to_float_vector vals.(3) in
  let maa = Uarray_ng.array_array_unpack lens a in
  report "initial coarse parameter extracted";
  debug (format_sprint print_coarse_parameter "maa" maa);
  let j = Typed_value.to_float vals.(4) in
  report "initial cost extracted";
  debug (format_sprint print_cost "j" j);
  (sz, ppm, maa, j)
;;

let make_init_wrapper =
  Pio.make_task_wrapper "init" mk_init_inj mk_init_proj
;;

(* Grad. *)

let mk_grad_inj { Udebug.spf_report = report; spf_debug = debug; _; } pm =
  report "sending one value";
  report "sending fine parameter";
  debug (format_sprint print_fine_parameter "pm" pm);
  [| Typed_value.of_c_float_matrix pm; |]
;;

let mk_grad_proj { Udebug.spf_report = report; spf_debug = debug; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._1_ then
    w_failwith "grad_proj" "gradient should return one value" else
  report "one value received";
  let gm = Typed_value.to_c_float_matrix vals.(0) in
  report "gradient matrix extracted";
  debug (format_sprint print_fine_gradient "gm" gm);
  gm
;;

let make_grad_wrapper =
  Pio.make_task_wrapper "grad" mk_grad_inj mk_grad_proj
;;

(* Optim. *)

let mk_optim_inj
    { Udebug.spf_report = report; spf_debug = debug; _; } (ppm, maa_init) =
  report "sending three values";
  report "sending partition";
  debug (format_sprint print_partition "ppm" ppm);
  let ppm = Umatrix.map Nat.to_int ppm in
  report "sending initial coarse parameter";
  debug (format_sprint print_coarse_parameter "maa_init" maa_init);
  let (lens_init, a_init) = Uarray_ng.array_array_pack maa_init in
  let lens_init = Array.map Nat.to_int lens_init in
  [| Typed_value.of_c_int_matrix ppm;
     Typed_value.of_int_vector lens_init;
     Typed_value.of_float_vector a_init; |]
;;

let mk_optim_proj { Udebug.spf_report = report; spf_debug = debug; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._3_ then
    w_failwith "optim_proj" "optimization should return three values" else
  report "three values received";
  let lens_opt = Typed_value.to_int_vector vals.(0) in
  let lens_opt = Array.map Nat.of_int lens_opt in
  report "optimal coarse parameter lengths extracted";
  let a_opt = Typed_value.to_float_vector vals.(1) in
  let maa_opt = Uarray_ng.array_array_unpack lens_opt a_opt in
  report "optimal coarse parameter extracted";
  debug (format_sprint print_coarse_parameter "maa_opt" maa_opt);
  let j_opt = Typed_value.to_float vals.(2) in
  report "optimal cost extracted";
  debug (format_sprint print_cost "j_opt" j_opt);
  (maa_opt, j_opt)
;;

let make_optim_wrapper =
  Pio.make_task_wrapper "optim" mk_optim_inj mk_optim_proj
;;

(* Final. *)

let mk_final_inj { Udebug.spf_report = report; spf_debug = debug; _; } pm =
  report "sending one value";
  report "sending fine parameter";
  debug (format_sprint print_fine_parameter "pm" pm);
  [| Typed_value.of_c_float_matrix pm; |]
;;

let mk_final_proj { Udebug.spf_report = report; spf_debug = debug; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._1_ then
    w_failwith "final_proj" "finalization should return one value" else
  report "one value received";
  let msg = Typed_value.to_string vals.(0) in
  debug (format_sprint Lib_print.print_quoted_string "msg" msg);
  if msg <> "OK, bye!" then
    w_failwith "final_proj" "finalization should return \"OK, bye!\"" else
  report "farewell message extracted";
;;

let make_final_wrapper =
  Pio.make_task_wrapper "final" mk_final_inj mk_final_proj
;;
